-module(functions).
-compile(export_all).

hello() -> io:format("Hello, world!~n").

% Defining functions with pattern matching
% Each of these function declarations is called a function clause.
% Function clauses must be separated by semicolons (;) 
% together form a function declaration. 
%
% the clauses are scanned sequentially,
% in the order in which they occur in the module,
% until the first one of them matches the call.
% When a match occurs,
% the expression on the right-hand side of the ‘->’ symbol is evaluated,
% and any variables occurring in the function definition are substituted
% in the right-hand side of the clause before it is evaluated.

greet(male, Name) ->
    io:format("Hello, Mr. ~s!~n", [Name]);
greet(female, Name) ->
    io:format("Hello, Mrs. ~s!~n", [Name]);
greet(_, Name) ->
    io:format("Hello, ~s!~n", [Name]).
% functions:greet(male,"Andrea").
% functions:greet(female,andrea) .
% functions:greet(boy,andrea) . 

% io:format alike printf
% In the string ~ denotes a token
% ~~ is ~ 
% ~n is line break
% ~s is a string or a bitstring
% ~p prints an erlang term in a nice way (with indentation)

% define second element of a list
second([_,X|_]) -> X.
% functions:second([a,b,c,d]).
% functions:second([a]).

% equality test via pattern matching
same(X,X) -> true;
same(_,_) -> false.
% functions:same(hello,hello).
% functions:same(hello,bye).

% it is possible to use the = operator in the function head, 
% allowing us to match both the content inside a tuple ({Y,M,D})
% and the tuple as a whole (Date)
valid_time({Date = {date,Y,M,D}, Time = {time,H,Min,S}}) ->
     io:format("The date tuple (~p) says today is: ~p/~p/~p,~n",[Date,Y,M,D]),
     io:format("The time tuple (~p) indicates: ~p:~p:~p.~n", [Time,H,Min,S]);
valid_time(_) ->
     io:format("Wrong data!~n").
% functions:valid_time({{date,2011,09,06},{time,09,04,43}}).
% functions:valid_time({{date,2011,09,06},{time,09,04}}).

% Guards are additional clauses to make pattern matching more expressive.
% guards will not accept user-defined functions (because of side effects)
old_enough(X) when X >= 16 -> true;
old_enough(_) -> false.
% functions:old_enough(15).

% , like andalso
right_age(X) when X >= 16, X =< 80 -> true;
right_age(_) -> false.
% functions:right_age(30).

% ; like orelse
wrong_age(X) when X < 16; X > 80 -> true;
wrong_age(_) -> false.
% functions:wrong_age(30).

% Note: , and ; are alike andalso and orelse, but NOT THE SAME.
% The former pair will catch exceptions as they happen while the latter won't.
% If there is an error thrown in the first part of the guard X >= N; N >= 0,
% the second part can still be evaluated and the guard might succeed;
% if an error was thrown in the first part of X >= N orelse N >= 0,
% the second part will also be skipped and the whole guard will fail.
% However only andalso and orelse can be nested inside guards.
% (A orelse B) andalso C is a valid guard, while (A; B), C is not.

% Ifs act like guards and share guards' syntax and restrictions,
% but outside of a function clause's head.
%
% Note: this one would be better as a pattern match in function heads!
% I'm doing it this way for the sake of the example.
help_me(Animal) ->
    Talk = if Animal == cat  -> "meow"
            ; Animal == beef -> "mooo"
            ; Animal == dog  -> "bark"
            ; true -> "gnagnagna"
           end,
    {Animal, "says " ++ Talk ++ "!"}.
% The body is a sequence of two statements, separated by a comma
% these statements are evaluated sequentially.
% The value of the clause is defined as the result of
% evaluating the last statement in the sequence.
% In the first statement of the sequence we introduced a temporary variable Talk.
% functions:help_me(cat).
% functions:help_me(monkey) .

% 'true' branches should be "avoided" altogether:
% ifs are usually easier to read when you cover all logical ends
% rather than relying on a "catch all" clause.

% we'll write the append function for sets (a collection of unique values)
% that we will represent as an unordered list.
% This is possibly the worst implementation possible in terms of efficiency,
% but what we want here is the syntax:
insert(X,[]) ->
    [X];
insert(X,Set) ->
    case lists:member(X,Set) of
         true  -> Set;
         false -> [X|Set]
    end.
% functions:insert(b,[a,b,c]).
% functions:insert(b,[a,c]).





% case of
beach(Temperature) ->
    case Temperature of
         {celsius, N} when N >= 20, N =< 45 -> 'favorable';
         {kelvin, N} when N >= 293, N =< 318 -> 'scientifically favorable';
         {fahrenheit, N} when N >= 68, N =< 113 -> 'favorable in the US';
         _ -> 'avoid beach'
    end.
% K = C + 273.15
% F = 32 + 1.8*C
% functions:beach({celsius,21}).

%%% Higher Order
one() -> 1.
two() -> 2.
% functions:one() .
% functions:two() .

add(X,Y) -> X() + Y().
% functions:add(fun functions:one/0 , fun functions:two/0).

apply(F,X) -> F(X).
% Succ = fun(X) -> X+1 end.
% functions:apply(Succ,3).
% lists:map(Succ,[1,2,3]).
